""" Configure and instantiate Celery
"""
import os

if os.environ.get('USE_GEVENT', "False").lower() == 'true':
    from gevent import monkey
    monkey.patch_all()

import sys
import logging
from pypeline.pipeline.chained_task import ChainedTask
from celery_dyrygent.tasks import register_workflow_processor
from typing import List
from celery import Celery
from pypeline.logging_config import setup_logging
from pypeline.utils.module_utils import SermosModuleLoader
from pypeline.utils.task_utils import PipelineResult, \
    get_service_config_for_worker
from pypeline.extensions import sermos_config, sermos_client_version
from pypeline import __version__

logger = logging.getLogger('celery')
ENABLE_TOOLS = str(os.environ.get('ENABLE_TOOLS', 'false')).lower() == 'true'
CELERY_TASKS_ACK_LATE = str(os.environ.get('CELERY_TASKS_ACK_LATE', 'false')).lower() == 'true'
LOG_LEVEL = os.environ.get("LOG_LEVEL", "INFO")
OVERLOAD_ES = os.environ.get('ENV', 'production').lower() == 'production'
PIPELINE_CHORD_COMPRESSION = os.environ.get('PIPELINE_CHORD_COMPRESSION', None)

setup_logging(app_version=__version__,
              client_version=sermos_client_version,
              default_level=LOG_LEVEL,
              overload_elasticsearch=OVERLOAD_ES,
              establish_logging_config=True)

def task_chain_regulator(*args, **kwargs):
    """ Utility task to ensure celery properly waits between groups in a chain.

        For a chain(), if each element is a group() then celery does not
        properly adhere to the chain elements occurring sequentially. If you
        insert a task that is not a group() in between, though, then the
        chain operates as expected.
    """
    return True


def pipeline_success(event: dict):
    """ Utility task to ensure celery properly waits between groups in a chain.

        For a chain(), if each element is a group() then celery does not
        properly adhere to the chain elements occurring sequentially. If you
        insert a task that is not a group() in between, though, then the
        chain operates as expected.
    """
    pr = PipelineResult(event['execution_id'])
    pr.load()
    pr.save(status='success')


class GenerateCeleryTasks(SermosModuleLoader):
    """ Use the sermos.yaml configuration to turn customer methods into
        decorated celery tasks that are available for work/pipelines
    """
    def __init__(self, config: dict, celery_instance: Celery):
        super(GenerateCeleryTasks, self).__init__()
        self.config = config if config else {}
        self.celery = celery_instance

    def _get_default_tasks(self) -> List[dict]:
        """ Sermos provides default tasks that all workers should know about.
        """
        return [{
            'handler': 'pypeline.celery.task_chain_regulator'
        }, {
            'handler': 'pypeline.celery.pipeline_success'
        }]

    def generate(self):
        """ Loads methods based on sermos config file and decorates them as
            celery tasks.

            Customer's methods:
            --------------------------------
            def demo_task(*args, **kwargs):
                return True

            Turns into the equivallent of:
            --------------------------------
            @celery.task(queue='queue-name')
            def demo_task(*args, **kwargs):t
                return True
        """
        # Set in k8s deployment as an environment variable when Sermos Cloud
        # generates the final secrets.yaml file. The name comes from the user's
        # sermos.yaml file based on serviceConfig[].name. Each 'worker' will
        # have a single name and each individually registers tasks through its
        # registeredTasks list. This allows each worker to only attempt
        # bootstrapping those tasks that are relevant to the worker and not, for
        # example, attempt to import a package that's not used by this worker
        service = get_service_config_for_worker(self.config)
        if not service:
            return
        for task in service.get('registeredTasks', []):
            pipeline_meta = None
            for pipeline_key, pipeline in sermos_config['pipelines'].items():
                pipeline_config = pipeline["config"]
                pipeline_tasks = [t["handler"] for t in pipeline_config["taskDefinitions"].values()]
                if task["handler"] in pipeline_tasks:
                    pipeline_meta = pipeline_config["metadata"]
                    break

            try:
                worker_path = task['handler']  # Required, no default

                tmp_handler = self.get_callable(worker_path)

                # Decorate the method as a celery task along with a default
                # queue if provided in config. Set ChainedTask as the base
                # which allows chained tasks to pass kwargs correctly.
                if pipeline_meta and pipeline_meta["maxRetry"] > 0:
                    tmp_handler = self.celery.task(
                        tmp_handler,
                        autoretry_for=(Exception,),
                        max_retries=pipeline_meta["maxRetry"],
                        retry_backoff=pipeline_meta["retryBackoff"],
                        retry_jitter=pipeline_meta["retryJitter"],
                        retry_backoff_max=pipeline_meta["retryBackoffMax"]
                    )
                else:
                    tmp_handler = self.celery.task(tmp_handler)
            except Exception as e:
                logger.warning(f"Unable to add a task to celery: {e}")
        # Sermos provides default tasks that all workers should know about, add
        # them here.
        for task in self._get_default_tasks():
            tmp_handler = self.get_callable(task['handler'])
            tmp_handler = self.celery.task(tmp_handler)


def configure_celery(celery: Celery):
    """ Configure Sermos-compatible Celery instance. Primarily this means
    compatibility with Pipelines and Scheduled Tasks through injecting the
    event kwarg. Also sets prebaked defaults that can be overloaded by user.
    """
    REDIS_URL = os.environ.get('REDIS_URL', 'redis://localhost:6379/0')
    CELERY_BROKER_URL = os.environ.get('CELERY_BROKER_URL', REDIS_URL)
    CELERY_RESULT_BACKEND = os.environ.get('CELERY_RESULT_BACKEND', REDIS_URL)

    celery.Task = ChainedTask

    # Configure the broker and tasks
    celery.conf.broker_url = CELERY_BROKER_URL

    # Use our custom database scheduler for dynamic celery beat updates.
    celery.conf.beat_scheduler =\
        'pypeline.celery_beat:SermosScheduler'

    # Reasonable defaults, override as necessary
    celery.conf.worker_redirect_stdouts = True
    celery.conf.worker_redirect_stdouts_level = LOG_LEVEL
    celery.conf.worker_hijack_root_logger = False

    if PIPELINE_CHORD_COMPRESSION:
      celery.conf.task_compression = PIPELINE_CHORD_COMPRESSION

    # NOTE: The broker URL may not be the best result backend. For example,
    # When using Rabbit as the broker (recommended), you should use Redis
    # as the result backend, as Rabbit has horrible support as backend.
    celery.conf.result_backend = CELERY_RESULT_BACKEND
    celery.conf.task_ignore_result = False  # Must not ignore for Chords
    celery.conf.result_expires = int(
        os.environ.get('CELERY_RESULT_EXPIRES', 10800))  # 3 hours by default
    celery.conf.broker_pool_limit = int(os.environ.get('BROKER_POOL_LIMIT',
                                                       10))
    celery.conf.worker_max_tasks_per_child = int(
        os.environ.get('MAX_TASKS_PER_CHILD', 100))
    celery.conf.task_soft_time_limit =\
        int(os.environ.get('TASK_TIMEOUT_SECONDS', 3600))
    celery.conf.task_time_limit =\
        int(os.environ.get('TASK_TIMEOUT_SECONDS', 3600)) + 10  # Cleanup buffer
    celery.conf.task_acks_late = CELERY_TASKS_ACK_LATE
    celery.conf.task_serializer = 'json'
    celery.conf.result_serializer = 'json'
    celery.conf.accept_content = ['json']
    # Required config options for some brokers we use frequently.
    transport_options = {}
    celery.conf.broker_transport_options = transport_options

    # Sermos generally has long-running tasks (relatively speaking), so
    # limit number of jobs a worker can reserve. This may not be true for
    # all tasks, so configure this on a per application basis. In the event
    # mutltiple task kinds exist in an application (short and long), see
    # http://docs.celeryproject.org/en/latest/userguide/optimizing.html#optimizing-prefetch-limit
    # for some guidance on combining multiple workers and routing tasks.
    # TODO make configurable from env
    celery.conf.worker_prefetch_multiplier = 1

    # Add our application's workers & any other tasks to be made
    # available
    register_workflow_processor(celery)
    try:
        GenerateCeleryTasks(sermos_config, celery).generate()
    except Exception as e:
        logger.error(f"Unable to dynamically generate celery tasks: {e}")
        sys.exit(1)

    return celery



""" Sermos implementation as a Flask extension
"""
import os
if os.getenv('USE_GEVENT', 'false').lower() == 'true':
    import gevent.monkey
    gevent.monkey.patch_all()

import logging
from flask import Flask
from werkzeug.middleware.proxy_fix import ProxyFix
from flask_smorest import  Api, Blueprint
from pypeline.extensions import sermos_config
from pypeline.constants import DEFAULT_OPENAPI_CONFIG
from pypeline import __version__

logger = logging.getLogger(__name__)


class FlaskSermos:
    """ Sermos Flask extension.
    """
    def __init__(self, app: Flask = None):
        """ Class init
        """
        self.app = app
        self.sermos_config = sermos_config if sermos_config is not None else {}

        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask, init_api: bool = False):
        """ Sermos bootstrapping process.

        Application config variables to set include:

            SERMOS_CLIENT_VERSION (default: v?.?.?)
            SERMOS_HIJACK_ROOT_LOGGER (default: False)

        Optional, if `init_api` is True:

            API_DOCUMENTATION_TITLE
            API_DOCUMENTATION_DESCRIPTION
            OPENAPI_VERSION
            OPENAPI_URL_PREFIX
            OPENAPI_SWAGGER_APP_NAME
            OPENAPI_SWAGGER_UI_PATH
            OPENAPI_SWAGGER_BASE_TEMPLATE
            OPENAPI_SWAGGER_URL
            OPENAPI_SWAGGER_UI_URL
            SWAGGER_UI_DOC_EXPANSION
            EXPLAIN_TEMPLATE_LOADING

        Args:
            app (Flask): Flask Application to initialize.
            init_api (bool): If `True`, Sermos will initialize its
                core APIs (including Pipelines, Scheduled Tasks, etc.) and
                provide a pre-configured OpenAPI Spec/Swagger UI interface
                available at the route defined in your application's config
                under `OPENAPI_URL_PREFIX` (default `/api`). Refer to
                [flask-smorest](https://flask-smorest.readthedocs.io/en/latest/openapi.html)
                documentation for additional configuration options.
        """
        # Ensure there's a SERMOS_CLIENT_VERSION on app config
        app.config.setdefault(
            'SERMOS_CLIENT_VERSION',
            app.config.get("SERMOS_CLIENT_VERSION", "v?.?.?"))

        app.wsgi_app = ProxyFix(app.wsgi_app)
        app.url_map.strict_slashes = False

        # Create and register the sermos blueprint
        bp = Blueprint('sermos',
                       __name__,
                       template_folder='../templates',
                       static_folder='../static',
                       url_prefix='/sermos')
        app.register_blueprint(bp)

        # Bootstrap api if app requests
        if init_api is True:
            self._bootstrap_api(app)

    def _bootstrap_api(self, app: Flask):
        """ If initializing the API, we will create the core Sermos API paths
            and initialize the default Swagger documentation.
        """
        # Set sensible defaults for Swagger docs. Provided `app` will
        # take precedent.
        for swagger_config in DEFAULT_OPENAPI_CONFIG:
            app.config.setdefault(
                swagger_config[0],
                app.config.get(swagger_config[0], swagger_config[1]))

        # Attempt to override with values from client's sermos.yaml if
        # they are available. This will add new tags and new docs if
        # defined and add to the core Sermos API docs.
        api_config = self.sermos_config.get('apiConfig', {})
        api_docs = api_config.get('apiDocumentation', {})

        custom_tags = api_config.get('prefixDescriptions', [])

        app.config['SERMOS_CLIENT_VERSION'] = \
            api_docs.get('version', None) \
            if api_docs.get('version', None) is not None \
            else app.config['SERMOS_CLIENT_VERSION']

        app.config['API_DOCUMENTATION_TITLE'] = \
            api_docs.get('title', None) \
            if api_docs.get('title', None) is not None \
            else app.config['API_DOCUMENTATION_TITLE']

        app.config['API_DOCUMENTATION_DESCRIPTION'] = \
            api_docs.get('description', None) \
            if api_docs.get('description', None) is not None \
            else app.config['API_DOCUMENTATION_DESCRIPTION']

        # Set default Sermos Tags along with custom tags from sermos.yaml
        tags = [{
            'name': 'Pipelines',
            'description': 'Operations related to Pipelines'
        }, {
            'name': 'Schedules',
            'description': 'Operations related to Schedules'
        }] + custom_tags

        # Set up  the initializing spec kwargs for API
        spec_kwargs = {
            'title': app.config['API_DOCUMENTATION_TITLE'],
            'version': f"Sermos: {__version__} - "
            f"Client: {app.config['SERMOS_CLIENT_VERSION']}",
            'description': app.config['API_DOCUMENTATION_DESCRIPTION'],
            'tags': tags
        }
        try:
            api = Api()
            api.init_app(app, spec_kwargs=spec_kwargs)

            # Register available Sermos API Namespaces
            self._register_api_namespaces(api)

        except Exception as e:
            api = None
            logging.exception(f"Unable to initialize API ... {e}")

        # Register the Sermos Core API as an extension for use in Client App
        app.extensions.setdefault('sermos_core_api', api)

    @staticmethod
    def _register_api_namespaces(api: Api):
        """ Register Default API namespaces
            TODO add metrics APIs
        """
        from pypeline.flask.api.pipelines import bp as pipelinesbp
        api.register_blueprint(pipelinesbp)
        from pypeline.flask.api.schedules import bp as schedulesbp
        api.register_blueprint(schedulesbp)



""" Sermos' Flask Implementation and Tooling. Convenience imports here.
"""
import logging

logger = logging.getLogger(__name__)

try:
    from flask_smorest import Blueprint, Api
    from flask import abort
except Exception as e:
    logger.error("Unable to import Web services (Blueprint, API, abort)"
                 f" ... {e}")

try:
    from pypeline.flask.flask_sermos import FlaskSermos
except Exception as e:
    logger.exception("Unable to import Sermos services (FlaskSermos)"
                     f" ... {e}")



""" Utilities for running and managing tasks inside pipelines.
"""
import logging

from celery import signature, chord, chain
from pypeline.utils.graph_utils import get_chainable_tasks
from pypeline.utils.config_utils import retrieve_latest_pipeline_config
from pypeline.utils.task_utils import PipelineRunWrapper, get_task_signature
from pypeline.constants import DEFAULT_TASK_TTL, DEFAULT_MAX_RETRY, \
    DEFAULT_REGULATOR_TASK, CHAIN_FAILURE_MSG, CHAIN_SUCCESS_MSG, \
    DEFAULT_SUCCESS_TASK
from pypeline.pipeline_config_schema import PipelineConfigValidator


logger = logging.getLogger(__name__)


class PipelineGenerator(object):
    """ Allows an API endpoint to generate a functional pipeline based on the
        requested pipeline id. Allows API to then issue the tasks asynchronously
        to initiate the pipeline. Thereafter, celery will monitor status and
        handle success/failure modes so the API web worker can return
        immediately.

        The primary purpose is to unpack the pipeline config, create the
        requisite cached entities to track pipeline progress, and apply the
        chained pipeline tasks asynchronously so Celery can take over.

        Usage:
            gen = PipelineGenerator(pipeline_id)
            chain = gen.generate_chain()
            chain.on_error(custom_error_task.s())  # Optional add error handling
            chain.delay()
    """
    def __init__(self,
                 pipeline_id: str,
                 access_key: str = None,
                 execution_id: str = None,
                 queue: str = None,
                 default_task_ttl: int = None,
                 regulator_queue: str = None,
                 regulator_task: str = None,
                 success_queue: str = None,
                 success_task: str = None,
                 default_max_retry: int = None,
                 retry_backoff: int = None,
                 retry_jitter: bool = None,
                 retry_backoff_max: int = None,
                 chain_payload: dict = None):
        super().__init__()
        self.pipeline_id = pipeline_id
        self.access_key = access_key

        pipeline_config_api_resp = retrieve_latest_pipeline_config(
            pipeline_id=self.pipeline_id, access_key=self.access_key)

        if pipeline_config_api_resp is None:
            raise ValueError("Unable to load Pipeline Configuration for "
                             f"pipeline id: {self.pipeline_id} ...")

        # The only part of the API response used for any 'pipeline config'
        # is the `config` key. The API nests it under `config` to preserve
        # ability to add additional detail at a later date.
        self.pipeline_config = pipeline_config_api_resp.get('config', {})
        schema_version = pipeline_config_api_resp.get('schemaVersion')
        PipelineConfigValidator(config_dict=self.pipeline_config,
                                schema_version=schema_version)

        self.execution_id = execution_id  # UUID string
        self.good_to_go = False  # Indicates initialization/loading success
        self.loading_message = None  # Allows access to success/error messages
        self.is_retry = False if self.execution_id is None else True

        self.default_max_retry = default_max_retry \
            if default_max_retry is not None else \
            self.pipeline_config['metadata'].get('maxRetry', DEFAULT_MAX_RETRY)
        self.retry_backoff = retry_backoff \
            if retry_backoff is not None else \
            self.pipeline_config['metadata'].get('retryBackoff', 3)
        self.retry_backoff_max = retry_backoff \
            if retry_backoff_max is not None else \
            self.pipeline_config['metadata'].get('retryBackoffMax', 600)
        self.retry_jitter = retry_jitter \
            if retry_jitter is not None else \
            self.pipeline_config['metadata'].get('retryJitter', False)

        # Queue on which to place tasks by default and default TTL per task
        # These can be overridden in PipelineConfig.config['taskDefinitions']
        self.queue = queue \
            if queue is not None \
            else self.pipeline_config['metadata']['queue']
        self.default_task_ttl = default_task_ttl \
            if default_task_ttl is not None else \
            self.pipeline_config['metadata'].get('maxTtl', DEFAULT_TASK_TTL)

        # See docstring in self._get_regulator()
        self.regulator_queue = regulator_queue \
            if regulator_queue is not None \
            else self.pipeline_config['metadata']['queue']
        self.regulator_task = regulator_task\
            if regulator_task is not None else DEFAULT_REGULATOR_TASK

        # See docstring in self._get_success_task()
        self.success_queue = success_queue \
            if success_queue is not None \
            else self.pipeline_config['metadata']['queue']
        self.success_task = success_task\
            if success_task is not None else DEFAULT_SUCCESS_TASK

        # Optional data to pass to each step in chain
        self.chain_payload = chain_payload\
            if chain_payload is not None else {}

        self.pipeline_wrapper = None  # Allows access to the PipelineRunWrapper
        self.chain = None  # Must be intentionally built with generate_chain()

        try:
            # Generate our wrapper for this pipeline_id / execution_id
            self.pipeline_wrapper = PipelineRunWrapper(
                pipeline_id=self.pipeline_id,
                pipeline_config=self.pipeline_config,
                execution_id=self.execution_id,
                max_ttl=self.default_task_ttl,
                max_retry=self.default_max_retry,
                chain_payload=self.chain_payload)

            # Loads pipeline config from remote or cache if it's already there
            # `is_retry` will be True for any PipelineGenerator instantiated
            # with an execution_id. This flag helps the wrapper increment the
            # retry count and determine if this should be deadlettered.
            # This step also saves the valid/initialized run wrapper to cache.
            self.pipeline_wrapper.load(is_retry=self.is_retry)

            # Set all variables that were established from the run wrapper
            # initialization. Notably, default_task_ttl can be overloaded
            # if the pipeline config has an explicit maxTtl set in metadata.
            self.good_to_go = self.pipeline_wrapper.good_to_go
            self.loading_message = self.pipeline_wrapper.loading_message
            self.execution_id = self.pipeline_wrapper.execution_id

        except Exception as e:
            fail_msg = "Failed to load Pipeline for id {} ... {}".format(
                self.pipeline_id, e)
            self.loading_message = fail_msg
            logger.error(fail_msg)
            raise e

    def _get_regulator(self):
        """ Create a chain regulator celery task signature.

            For a chain(), if each element is a group() then celery does not
            properly adhere to the chain elements occurring sequentially. If you
            insert a task that is not a group() in between, though, then the
            chain operates as expected.
        """
        return signature(self.regulator_task,
                         queue=self.regulator_queue,
                         immutable=True)

    def _get_success_task(self):
        """ A final 'success' task that's added to the end of every pipeline.

            This stores the 'success' state in the cached result. Users can
            set other values by using TaskRunner().save_result()
        """
        return get_task_signature(task_path=self.success_task,
                                  queue=self.success_queue,
                                  pipeline_id=self.pipeline_id,
                                  execution_id=self.execution_id)

    def _get_signature(self, node):
        """ Create a celery task signature based on a graph node.
        """
        metadata = self.pipeline_config['metadata']
        node_config = self.pipeline_config['taskDefinitions'][node]

        # Node config takes precedence, pipeline metadata as default
        queue = node_config.get('queue', metadata['queue'])
        max_ttl = node_config.get('maxTtl', metadata.get('maxTtl', None))

        # Ensures task signatures include requisite information to retrieve
        # PipelineRunWrapper from cache using the pipeline id, and execution id.
        # We set immutable=True to ensure each client task can be defined
        # with this specific signature (event)
        # http://docs.celeryproject.org/en/master/userguide/canvas.html#immutability
        return get_task_signature(task_path=node_config.get('handler'),
                                  queue=queue,
                                  access_key=self.access_key,
                                  pipeline_id=self.pipeline_id,
                                  execution_id=self.execution_id,
                                  max_ttl=max_ttl,
                                  immutable=True,
                                  task_config=node_config)

    def generate_chain(self):
        """ Generate the full pipeline chain.
        """
        logger.debug(f'Starting Pipeline {self.pipeline_id}')

        if not self.good_to_go:
            logger.info("Chain deemed to be not good to go.")
            if self.loading_message is None:
                self.loading_message = CHAIN_FAILURE_MSG
            return None

        try:
            # Create the task chain such that all concurrent tasks are grouped
            # and all high level node groups are run serially
            G = self.pipeline_wrapper.execution_graph

            total_tasks = 0
            pipeline_chain = []
            chainable_tasks = get_chainable_tasks(G, None, [])

            # Current chord+chain solution based on
            # https://stackoverflow.com/questions/15123772/celery-chaining-groups-and-subtasks-out-of-order-execution
            # Look also at last comment from Nov 7, 2017 here
            # https://github.com/celery/celery/issues/3597
            # Big outstanding bug in Celery related to failures in chords that
            # results in really nasty log output. See
            # https://github.com/celery/celery/issues/4834
            for i, node_group in enumerate(chainable_tasks):
                total_tasks += len(node_group)
                this_group = []
                for node in node_group:
                    node_signature = self._get_signature(node)
                    this_group.append(node_signature)

                if len(this_group) <= 1:
                    this_group.append(self._get_regulator())

                the_chord = chord(header=this_group,
                                  body=self._get_regulator())

                pipeline_chain.append(the_chord)

            # Add a 'finished/success' task to the end of all pipelines
            pipeline_chain.append(
                chord(header=self._get_success_task(),
                      body=self._get_regulator()))

            the_chain = chain(*pipeline_chain)

            self.loading_message = CHAIN_SUCCESS_MSG

            self.chain = the_chain
        except Exception as e:
            self.loading_message = CHAIN_FAILURE_MSG + " {}".format(e)
            logger.exception(e)
            the_chain = None

        self.chain = the_chain

        return the_chain



import logging
from celery import Task
from pypeline.utils.task_utils import PipelineRunWrapper


logger = logging.getLogger(__name__)


class ChainedTask(Task):
    """ A Celery Task that is used as the _base_ for all dynamically
    generated tasks (by GenerateCeleryTasks().generate()). This injects
    `event` into every task's signature, which allows pipelines to pass
    event information easily through a chain.
    """
    abstract = True

    def __call__(self, *args, **kwargs):
        """ Allow the return value of one task to update the kwargs of a
            subsequent task if it's a dictionary. Important to the function
            of a pipeline to allow event information to flow easily.
        """
        # Inject app context
        if len(args) == 1 and isinstance(args[0], dict):
            kwargs.update(args[0])
            args = ()

        # Event holds information used in PipelineRunWrapper and
        # other areas.
        if 'event' not in kwargs.keys():
            kwargs['event'] = {}
        # This is a special worker from dyrygent that orchestrates our
        # pipelines.  It provides a patch in fix for celery's poor
        # implementation of Canvas work-flows
        if self.__name__ == 'workflow_processor':
            kwargs.pop('event', None)
        return super(ChainedTask, self).__call__(*args, **kwargs)

    def after_return(self, status, retval, task_id, args, kwargs, einfo):
        if "event" in kwargs and "pipeline_id" in kwargs["event"]:
            try:
                pipeline_run_wrapper: PipelineRunWrapper = \
                    PipelineRunWrapper.from_event(kwargs["event"])
                current_task_status = pipeline_run_wrapper.get_task_celery_status(task_id)
            except Exception:
                logger.exception("Unable to retreive Pipeline Run Wrapper")
                return

            if current_task_status:
                current_task_status["status"] = status
            try:
                pipeline_run_wrapper.save_to_cache()
            except Exception:
                logger.exception(f"Failed to update celery task status for task {task_id}")

    def on_retry(self, exc, task_id, args, kwargs, einfo):
        if "event" in kwargs and "pipeline_id" in kwargs["event"]:
            try:
                pipeline_run_wrapper: PipelineRunWrapper = \
                    PipelineRunWrapper.from_event(kwargs["event"])
                current_task_status = pipeline_run_wrapper.get_task_celery_status(task_id)
            except Exception:
                logger.exception("Unable to retreive Pipeline Run Wrapper")
                return

            if current_task_status:
                current_task_status["retries"] = current_task_status["retries"] + 1
            try:
                pipeline_run_wrapper.save_to_cache()
            except Exception:
                logger.exception(f"Failed to update celery task status for task {task_id}")



""" PypeLine Library Setup
"""
import re
import ast
import sys
import os
from typing import List

from setuptools import setup, find_packages, Extension
from setuptools.command.build_py import build_py
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    """ PyTest Command
    """

    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ''

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import coverage
        import pytest

        if self.pytest_args and len(self.pytest_args) > 0:
            self.test_args.extend(self.pytest_args.strip().split(' '))
            self.test_args.append('tests/')

        cov = coverage.Coverage()
        cov.start()
        errno = pytest.main(self.test_args)
        cov.stop()
        cov.report()
        cov.html_report()
        print("Wrote coverage report to htmlcov directory")
        sys.exit(errno)


# Defaults that will be overridden/updated if do_cythonize is true.
#
ext_modules = []
cmdclass = {'test': PyTest}
packages = find_packages(exclude=["tests"])

do_cythonize = os.getenv('CYTHONIZE', 'false').lower() == 'true'
if do_cythonize:
    try:
        from Cython.Build import cythonize
        from Cython.Distutils import build_ext

        cmdclass['build_ext'] = build_ext  # Use cython's build_ext

        def scandir(directory: str,
                    exclude_paths: List[str] = None,
                    paths: List[str] = None,
                    directories: List[str] = None,
                    return_directories: bool = False,
                    recursive: bool = True):
            """ Scan a given directory recursively and produce a list
                of file paths (e.g. ['pypeline/app.py', 'pypeline/config.py'])
                or directroies (e.g. ['pypeline/api', 'pypeline/utils'])
            """
            if paths is None:
                paths = []
            if directories is None:
                directories = []
            if exclude_paths is None:
                exclude_paths = []

            for file in os.listdir(directory):
                path = os.path.join(directory, file)
                if any([p in path for p in exclude_paths]):
                    continue

                if '__init__.py' in path or '__pycache__' in path:
                    continue

                if os.path.isfile(path) and path.endswith(".py"):
                    paths.append(path)
                elif os.path.isdir(path) and recursive:
                    directories.append(path)
                    scandir(path,
                            exclude_paths=exclude_paths,
                            paths=paths,
                            directories=directories,
                            return_directories=return_directories,
                            recursive=recursive)

            if return_directories:
                return directories
            return paths

        def make_extension(ext_path: str):
            """ Generate an Extension() object. Takes a path
                e.g. pypeline/app.py
                and generates a valid Extension
                e.g. Extension('pypeline.app', ['pypeline/app.py'])
            """
            extName = ext_path.replace("/", ".")[:-3]
            return Extension(extName, [ext_path], include_dirs=["."])

        to_cythonize = scandir('pypeline',
                               exclude_paths=[
                                   'pypeline/templates', 'pypeline/static',
                                   'pypeline/celery.py',
                                   'pypeline/tools/thumbnail/thumbnail.py',
                                   'pypeline/lib/config_server.py'
                               ])

        ext_modules = cythonize(
            [make_extension(path) for path in to_cythonize], language_level=3)

        class MyBuildPy(build_py):
            """ Update standard build_py to exclude any files we
                explicitly cythonize.
            """
            def find_package_modules(self, package, package_dir):
                """ Return list of all package modules that are *not* cythonized
                    for regular packaging.
                """
                modules = super().find_package_modules(package, package_dir)
                filtered_modules = []
                for tup in modules:
                    if tup[2] in to_cythonize:
                        continue
                    filtered_modules.append(tup)
                return filtered_modules

        cmdclass['build_py'] = MyBuildPy

    except ImportError:
        pass

_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('pypeline/__init__.py', 'rb') as f:
    __version__ = str(
        ast.literal_eval(
            _version_re.search(f.read().decode('utf-8')).group(1)))

with open('requirements.txt', 'r') as f:
    install_requires = f.read().splitlines()

setup(
    name='scalable-pypeline',
    version=__version__,
    description="PypeLine - Python pipelines for the Real World",
    long_description=open('README.md', 'r').read(),
    long_description_content_type="text/markdown",
    author="Bravos Power Corporation",
    license="Apache License 2.0",
    url="https://gitlab.com/bravos2/pypeline",
    packages=packages,
    include_package_data=True,
    cmdclass=cmdclass,
    ext_modules=ext_modules,
    install_requires=install_requires,
    extras_require={
        'build': ['wheel', 'twine'],
        'flask': [
            'Werkzeug==2.0.3',
            'Flask>=1.1.2,<2',
            'flask-smorest>=0.23.0,<0.29',
            'Jinja2==3.0.3',
        ],
        'web': ['gunicorn', 'gevent>=21.12.0,<22'],
        'workers': [
            "celery[redis]>=5.1.2,<6",
            'networkx>=2.4',
        ],
        'dev':['honcho>=1.0.1', 'awscli>=1.11', 'pylint>=2.5.3', 'pip-licenses'],
        'test': [
            'pytest-cov>=2.6.1,<3',
            'tox>=3.14.1,<4',
            'mock>=1,<2',
            'moto>=1.3.16',
            'responses>=0.10.16,<0.11',
            'fakeredis>=2.10.3,<3',
            'importlib-metadata>=4.12,<5'
        ]
    },
    entry_points="""
    [console_scripts]
    pypeline=pypeline.cli.core:pypeline
    """,
)
